from pycoder.api.main import query
from pycoder.utils import formatter
from pycoder.version import __version__
from pycoder.imports import (
    time,
    List,
    echo,
    Exit,
    Typer,
    Option,
    Optional,
    spinner,
    highlight,
    PythonLexer,
    TerminalFormatter,
    system,
)

app = Typer()


def version_callback(value: bool):
    if value:
        echo(f"Pycoder 🐍 CLI version: {__version__}")
        raise Exit()


############################
#   interface for inference
############################


def endpoint_callback(value: bool):
    if value:
        from pycoder.api.app import main

        main()
        raise Exit()


@app.command()
def main(
    topic: List[str] = Option(
        ..., "--topic", "-t", help="topic[s] for the code to be written."
    ),
    description: str = Option(
        ..., "--description", "-d", help="description of the code to be written."
    ),
    prefix: Optional[str] = Option(
        "", "--prefix", "-p", help="starting of the code for pycoder to follow."
    ),
    max_length: Optional[int] = Option(
        200, "--max-length", "-ml", help="max length of code to generate."
    ),
    cuda: Optional[bool] = Option(
        False,
        "--cuda",
        "-c",
        help="if cuda device is to be used for inference or not. Might not work in case VRAM is <4GB.",
    ),
    execution_time: Optional[bool] = Option(
        False,
        "--execution_time",
        "-et",
        help="get execution time of generating the code.",
    ),
    endpoint: Optional[bool] = Option(
        None,
        "--endpoint",
        "-e",
        callback=endpoint_callback,
        help="specify the port to start the endpoint on.",
    ),
    version: Optional[bool] = Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        help="shows the current version of Pycoder.",
    ),
) -> str:
    """
    CLI command to get Python 🐍 code from set of topic[s] and description.
    Examples:\n
    🔥 pycoder --topic pytorch --description "a trainer for vision"\n
    🔥 pycoder -t pytorch -d "a trainer for vision"\n
    🔥 pycoder --topic pytorch --topic torch --description "a trainer for vision"\n
    🔥 pycoder -t pytorch -t torch -d "a trainer for vision"\n
    🔥 pycoder --topic pytorch --topic torch --description "a trainer for vision" --prefix "class Trainer:"\n
    🔥 pycoder -t pytorch -t torch -d "a trainer for vision" --prefix "class Trainer:"\n
    """
    st = time()
    with spinner():
        code = query(list(topic), description, prefix, max_length, cuda, False)
    et = time()

    echo(highlight(code, PythonLexer(), TerminalFormatter()))

    if execution_time:
        echo(formatter(f"took {(et-st):.2f} seconds to code 😀", color="g", bold=True))


def cli():
    app()



from pycoder.utils import *
from pycoder.imports import system, time, json, List, Union, tqdm, Path, datetime

try:
    from pycoder.imports import Github, Repository
except:
    pass


def repo_to_dict(repository: Repository, keyword: str) -> dict:
    try:
        return {
            "keyword": keyword,
            "url": repository.clone_url,
            "owner": repository.owner.login,
            "name": repository.name,
            "topics": repository.get_topics(),
            "readme_url": repository.get_readme().download_url,
            "language": repository.language,
        }
    except:
        return None


def collect_repository(
    g: Github,
    keyword: str,
    language: str = "python",
    num_instances: int = 400,
) -> List[dict]:
    end_time = time()
    start_time = end_time - 86400  # a day ago

    repositories = []

    while num_instances > 0:
        print(f"Collecting {formatter(keyword.capitalize(), color='g', bold=True)}...")

        start_time_str = datetime.utcfromtimestamp(start_time).strftime("%Y-%m-%d")
        end_time_str = datetime.utcfromtimestamp(end_time).strftime("%Y-%m-%d")
        not_added = 0

        query = (
            f"{keyword} language:{language} created:{start_time_str}..{end_time_str}"
        )

        results = g.search_repositories(query)

        for repository in tqdm(results, total=results.totalCount):
            repository_json = repo_to_dict(repository, keyword)
            if repository_json:
                repositories.append(repository_json)
            else:
                not_added += 1

        start_time -= 86400
        end_time -= 86400
        num_instances -= results.totalCount
        num_instances += not_added

    return repositories


def download_github_repositories(
    save_path: Union[Path, str],
    github_access_token: str,
    keywords: List[str],
) -> None:
    save_path = Path(save_path)
    save_path.parent.mkdir(parents=True, exist_ok=True)

    g = Github(github_access_token)

    repositories = []
    for keyword in keywords:
        repositories += collect_repository(g, keyword)

    print(f"Saving to {formatter(str(save_path), color='g', bold=True)}...")
    with open(str(save_path), "w") as fl:
        json.dump(repositories, fl)


def clone_repository(repository_dict: dict, save_path: Union[Path, str]) -> None:
    save_path = Path(save_path) / repository_dict["keyword"]
    save_path.mkdir(exist_ok=True, parents=True)

    url = repository_dict["url"]
    owner = repository_dict["owner"].lower()
    name = repository_dict["name"].lower()

    cmd = f"git clone {url} {save_path /(owner+'___'+name)}"

    system(cmd)


def clone_repositories(
    json_path: Union[Path, str], save_path: Union[Path, str]
) -> None:
    with open(json_path, "r") as fl:
        repositories = json.load(fl)

    repositories = tqdm(repositories)
    for repository in repositories:
        repositories.set_description(f"Cloning {repository['name']}")
        clone_repository(repository, save_path)



from pycoder.version import __version__
from pycoder.imports import getenv, Path

try:
    from pycoder.imports import load_dotenv

    load_dotenv()
except:
    pass

#################
# secrets
##################
GITHUB_ACCESS_TOKEN = getenv("GITHUB_ACCESS_TOKEN")
WANDB_API_KEY = getenv("WANDB_API_KEY")


####################
# configuration
####################

# model config
GPT_CONFIGS = {
    "S": {"MODEL_NAME": "gpt2", "MAX_LENGTH": 768},
    "M": {"MODEL_NAME": "gpt2-medium", "MAX_LENGTH": 1024},
    "L": {"MODEL_NAME": "gpt2-large", "MAX_LENGTH": 1280},
    "XL": {"MODEL_NAME": "gpt2-xl", "MAX_LENGTH": 1600},
}

HF_HUB_NAME = "himanshu-dutta/pycoder-gpt2"
MODEL_NAME = GPT_CONFIGS["M"]["MODEL_NAME"]
MAX_LENGTH = GPT_CONFIGS["M"]["MAX_LENGTH"]
NUM_DESCRIPTION_SENTENCES = 2

SPECIAL_TOKENS = {
    "bos_token": "<|BOS|>",
    "eos_token": "<|EOS|>",
    "unk_token": "<|UNK|>",
    "pad_token": "<|PAD|>",
    "sep_token": "<|SEP|>",
}
CONTROL_TOKENS = {
    "topics_token": "<|TOP|>",
    "description_token": "<|DESC|>",
    "code_token": "<|CODE|>",
    "eos_token": SPECIAL_TOKENS["eos_token"],
}

# training parameters
EPOCHS = 5
TRAIN_BATCHSIZE = 1
BATCH_UPDATE = 16
LR = 5e-4
EPS = 1e-8
WARMUP_STEPS = 1e2
APEX_OPT_LEVEL = "01"
FP16 = True
VAL_PCT = 0.2
RUN_NAME = f"RUN:{(__version__.split('.'))[-1]}"

# paths and directories
PACKAGE_DIR = Path(__file__).parent.absolute()
ARTIFACT_PATH = PACKAGE_DIR / "artifacts"

CACHE_DIR = PACKAGE_DIR / "cache"

ASSETS_DIR = PACKAGE_DIR / "assets"
CONFIG_PATH = Path(__file__)

DATA_DIR = ARTIFACT_PATH / "data"
REPO_DATA_PATH = DATA_DIR / "repositories"
README_DATA_PATH = DATA_DIR / "readmes"

MODEL_DIR = ARTIFACT_PATH / "models"
MODEL_PATH = MODEL_DIR / RUN_NAME / MODEL_NAME / "model"
TOKENIZER_PATH = MODEL_DIR / RUN_NAME / MODEL_NAME / "tokenizer"

CHECKPOINT_PATH = ARTIFACT_PATH / "checkpoints" / RUN_NAME / MODEL_NAME

REPO_JSON_PATH = DATA_DIR / "repositories.json"
REPO_INDEX_JSON_PATH = DATA_DIR / "indexed_repos.json"
FILE_INDEX_JSON_PATH = DATA_DIR / "indexed_files.json"

KEYWORDS = ["tensorflow", "pytorch", "django", "flask", "algorithms", "fastapi"]

#############
# constants
#############

TICK = "✅"
CROSS = "❌"

BOLD = "\033[1m"
GREEN = "\033[32m"
RED = "\033[91m"
END = "\033[0;0m"

GITIGNORES = [
    "__pycache__/",
    "build/",
    "develop-eggs/",
    "dist/",
    "downloads/",
    "eggs/",
    ".eggs/",
    "lib/",
    "lib64/",
    "parts/",
    "sdist/",
    "var/",
    "wheels/",
    "share/python-wheels/",
    ".egg-info/",
    "htmlcov/",
    ".tox/",
    ".nox/",
    ".hypothesis/",
    ".pytest_cache/",
    "cover/",
    "instance/",
    "docs/_build/",
    ".pybuilder/",
    "target/",
    ".ipynb_checkpoints",
    "profile_default/",
    "ipython_config.py",
    "__pypackages__/",
    ".sage.py",
    ".env",
    ".venv",
    "env/",
    "venv/",
    "ENV/",
    "env.bak/",
    "venv.bak/",
    ".spyderproject",
    ".spyproject",
    ".ropeproject",
    "/site",
    ".mypy_cache/",
    ".dmypy.json",
    "dmypy.json",
    ".pyre/",
    ".pytype/",
    "cython_debug/",
]



from pycoder.utils import *
from pycoder.imports import (
    json,
    AutoConfig,
    AutoModelForPreTraining,
    AutoTokenizer,
    Union,
    Path,
    Tuple,
)


def get_tokenier(
    model_name_or_path: Union[Path, str] = None,
    special_tokens: dict = None,
) -> "transformer.PreTrainedTokenizer":

    tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)

    if special_tokens:
        tokenizer.add_special_tokens(special_tokens)

    return tokenizer


def get_model(
    model_name_or_path: Union[Path, str],
    tokenizer: "transformer.PreTrainedTokenizer",
    special_tokens=None,
) -> "nn.Module":

    if special_tokens:
        config = AutoConfig.from_pretrained(
            model_name_or_path,
            bos_token_id=tokenizer.bos_token_id,
            eos_token_id=tokenizer.eos_token_id,
            sep_token_id=tokenizer.sep_token_id,
            pad_token_id=tokenizer.pad_token_id,
            output_hidden_states=False,
        )
    else:
        config = AutoConfig.from_pretrained(
            model_name_or_path,
            pad_token_id=tokenizer.eos_token_id,
            output_hidden_states=False,
        )

    model = AutoModelForPreTraining.from_pretrained(model_name_or_path, config=config)

    if special_tokens:
        model.resize_token_embeddings(len(tokenizer))

    return model


def save_transformers(
    model_path: Union[Path, str],
    tokenizer_path: Union[Path, str],
    model: "nn.Module",
    tokenizer: "transformer.PreTrainedTokenizer",
    verbose: bool = True,
) -> None:
    model_path, tokenizer_path = Path(model_path), Path(tokenizer_path)

    model_path.mkdir(parents=True, exist_ok=True)
    tokenizer_path.mkdir(parents=True, exist_ok=True)

    with (tokenizer_path / "config.json").open("w") as fl:
        json.dump({"model_type": "gpt2"}, fl)

    model.save_pretrained(model_path)
    tokenizer.save_pretrained(tokenizer_path)

    if verbose:
        print(
            "Saved the model to ",
            formatter(str(model_path), color="g", bold=True, tick=True),
        )
        print(
            "Saved the tokenizer to ",
            formatter(str(tokenizer_path), color="g", bold=True, tick=True),
        )


def load_transformers(
    model_path: Union[Path, str] = None,
    tokenizer_path: Union[Path, str] = None,
    verbose: bool = False,
) -> Tuple["nn.Module", "transformer.PreTrainedTokenizer"]:

    model_path, tokenizer_path = Path(model_path), Path(tokenizer_path)

    assert (
        model_path.exists() and tokenizer_path.exists()
    ), "Both model and tokenizer path must exist."

    tokenizer = get_tokenier(tokenizer_path)
    model = get_model(model_path, tokenizer)

    if verbose:
        print(
            "Loaded the model from ",
            formatter(str(model_path), color="g", bold=True, tick=True),
        )
        print(
            "Loaded the tokenizer from ",
            formatter(str(tokenizer_path), color="g", bold=True, tick=True),
        )

    return model, tokenizer



from pycoder.model.inference import CodeInference
from pycoder.imports import Union, Path, List, lru_cache, Path


MODEL_PATH = Path(__file__).parent.parent / "assets" / "model"
TOKENIZER_PATH = Path(__file__).parent.parent / "assets" / "tokenizer"


@lru_cache(1)
def create_inference_instance(
    model_path: Union[Path, str] = MODEL_PATH,
    tokenizer_path: Union[Path, str] = TOKENIZER_PATH,
    max_length: int = 200,
    cuda: bool = False,
    verbose: bool = False,
) -> CodeInference:
    CONTROL_TOKENS = {
        "topics_token": "<|TOP|>",
        "description_token": "<|DESC|>",
        "code_token": "<|CODE|>",
        "eos_token": "<|EOS|>",
    }

    return CodeInference(
        model_path,
        tokenizer_path,
        CONTROL_TOKENS,
        max_length,
        cuda,
        verbose,
    )


def query(
    topics: Union[List[str], str],
    description: str,
    code_prefix: str = "",
    max_length: int = 200,
    cuda: bool = False,
    verbose: bool = False,
) -> str:

    coder = create_inference_instance(max_length=max_length, cuda=cuda, verbose=verbose)

    out = coder(topics, description, code_prefix)

    if len(out) > 0:
        return out[0]

    return None



from pycoder.data.classes import Code
from pycoder.imports import Dataset, List, Dict
from pycoder.data.processing import sentence_tokenize_and_select_random


class CodeDataset(Dataset):
    def __init__(
        self,
        codes: List[Code],
        tokenizer: "transformer.PreTrainedTokenizer",
        control_tokens: dict,
        max_length: int,
        num_description_sentences: int,
    ) -> None:
        self.codes = codes
        self.tokenizer = tokenizer
        self.control_tokens = control_tokens
        self.max_length = max_length
        self.num_description_sentences = num_description_sentences

    def get_formatted_code(self, index: int) -> str:
        """
        returns a string with the format:
        <|TOP|>TOPICS<|DES|>DESCRIPTION<|CODE|>CODE<|EOS|>
        """
        code = self.codes[index]
        content = code.content
        description = sentence_tokenize_and_select_random(
            code.readme, self.num_description_sentences
        )
        topics = ",".join(code.topics) if len(code.topics) else ""
        topics = code.category + "," + topics

        return (
            self.control_tokens["topics_token"]
            + topics
            + self.control_tokens["description_token"]
            + description
            + self.control_tokens["code_token"]
            + content
            + self.control_tokens["eos_token"]
        )

    def __len__(self) -> int:
        return len(self.codes)

    def __getitem__(self, index: int) -> Dict[str, "torch.tensor"]:
        formatted_code = self.get_formatted_code(index)

        tokenized = self.tokenizer(
            formatted_code,
            truncation=True,
            max_length=self.max_length,
            padding="max_length",
            return_tensors="pt",
        )

        return {
            "label": tokenized["input_ids"],
            "input_ids": tokenized["input_ids"],
            "attention_mask": tokenized["attention_mask"],
        }

    def get_string(self, index: int) -> str:
        formatted_code = self.get_formatted_code(index)
        return formatted_code



import pycoder.config as cfg
import pycoder.utils as utils
from pycoder.model import training
from pycoder.data import download, processing
from pycoder.imports import List, Typer, system

app = Typer()


###################
#   data commands
###################


@app.command()
def download_github_repositories(
    save_path: str = cfg.REPO_JSON_PATH,
    github_accesss_token: str = cfg.GITHUB_ACCESS_TOKEN,
    keywords: List[str] = cfg.KEYWORDS,
):
    download.download_github_repositories(save_path, github_accesss_token, keywords)


@app.command()
def clone_repositories(
    json_path: str = cfg.REPO_JSON_PATH, save_path: str = cfg.REPO_DATA_PATH
):
    download.clone_repositories(json_path, save_path)


@app.command()
def index_repositories(
    json_path: str = cfg.REPO_JSON_PATH, save_path: str = cfg.REPO_INDEX_JSON_PATH
):
    processing.index_repositories(json_path, save_path)


@app.command()
def clean_repositories(
    root_dir: str = cfg.REPO_DATA_PATH,
    must_remove_phrases: List[str] = cfg.GITIGNORES,
):
    processing.walk_clean(root_dir, must_remove_phrases)


@app.command()
def index_files(
    root_dir: str = cfg.REPO_DATA_PATH,
    save_path: str = cfg.FILE_INDEX_JSON_PATH,
    readme_save_path: str = cfg.README_DATA_PATH,
    repository_index: str = cfg.REPO_INDEX_JSON_PATH,
    keywords: List[str] = cfg.KEYWORDS,
):
    processing.index_files_from_root(
        root_dir,
        save_path,
        readme_save_path,
        repository_index,
        keywords,
    )


@app.command()
def get_data(
    repository_json_path: str = cfg.REPO_JSON_PATH,
    github_accesss_token: str = cfg.GITHUB_ACCESS_TOKEN,
    search_keywords: List[str] = cfg.KEYWORDS,
    must_remove_phrases: List[str] = cfg.GITIGNORES,
    repository_data_path: str = cfg.REPO_DATA_PATH,
    repository_index_json_path: str = cfg.REPO_INDEX_JSON_PATH,
    file_index_json_path: str = cfg.FILE_INDEX_JSON_PATH,
    readme_save_path: str = cfg.README_DATA_PATH,
):
    # download github repository data to a json file
    download.download_github_repositories(
        repository_json_path,
        github_accesss_token,
        search_keywords,
    )

    # clone the repository to a local folder
    download.clone_repositories(
        repository_json_path,
        repository_data_path,
    )

    # index the repositories to a json file
    processing.index_repositories(repository_json_path, repository_index_json_path)

    # clean the repositories for any unwanted file
    processing.walk_clean(repository_data_path, must_remove_phrases)

    # index the required files and download the readme for them
    processing.index_files_from_root(
        repository_data_path,
        file_index_json_path,
        readme_save_path,
        repository_index_json_path,
        search_keywords,
    )


###################
#   model commands
###################


@app.command()
def run_training(checkpoint_dir: str = None):
    training.run_training(cfg, checkpoint_dir)


#####################
#   utility commands
#####################


@app.command()
def dump_assets(assets_dir: str = None, cfg=cfg):
    if assets_dir == None:
        assert cfg.ASSETS_DIR != None, "assets_dir / cfg.ASSETS_DIR is required."
        assets_dir = str(cfg.ASSETS_DIR)

    assert (
        cfg.MODEL_PATH != None
        and cfg.TOKENIZER_PATH != None
        and cfg.CONFIG_PATH != None
    )

    assets = [str(cfg.MODEL_PATH), str(cfg.TOKENIZER_PATH), str(cfg.CONFIG_PATH)]

    for asset in assets:
        cmd = f"cp -r {asset} {assets_dir}"
        system(cmd)
        print(utils.formatter(f"dumped {asset}", color="g", bold=True, tick=True))


@app.command()
def update_version():
    utils.update_version()


if __name__ == "__main__":
    app()



######################
#   python imports
######################

import re
import json
import codecs
import requests
from time import time
from io import StringIO
from pathlib import Path
from shutil import rmtree
from random import choices
from markdown import Markdown
from datetime import datetime
from functools import lru_cache
from dataclasses import dataclass, field
from typing import List, Union, Dict, Tuple, Optional
from os import getenv, system, path, walk, listdir


######################
#   model imports
######################

import torch
from torch.utils.data import DataLoader, Dataset, random_split
from transformers import (
    AutoTokenizer,
    AutoConfig,
    GPT2LMHeadModel,
    AutoModelForPreTraining,
    AdamW,
    TrainingArguments,
    BeamScorer,
    Trainer,
    get_linear_schedule_with_warmup,
    pipeline,
)


######################
#   api imports
######################

from tqdm import tqdm
from typer import Typer, Option, Exit, echo

from uvicorn import run
from fastapi import FastAPI
from asyncio import create_task
from click_spinner import spinner

from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import TerminalFormatter


######################
#   dev imports
######################

try:
    from github import Github
    from github.Repository import Repository
    from dotenv import load_dotenv
except:
    pass



from pycoder.utils import formatter
from pycoder.imports import (
    walk,
    path,
    listdir,
    re,
    json,
    rmtree,
    requests,
    tqdm,
    StringIO,
    Path,
    Markdown,
    Dict,
    List,
    Union,
    choices,
)


def index_repositories(
    json_path: Union[Path, str], save_path: Union[Path, str]
) -> None:
    with Path(json_path).open("r") as fl:
        repositories = json.load(fl)

    repository_dict = dict()
    for repository in repositories:
        owner = repository["owner"].lower()
        name = repository["name"].lower()
        repository_dict[f"{owner}___{name}"] = repository

    with Path(save_path).open("w") as fl:
        json.dump(repository_dict, fl)


def walk_clean(
    root_dir: Union[Path, str],
    must_remove_phrases: List[str],
    keep_extensions: List[str] = ["py"],
) -> None:
    dir_walk = tqdm(walk(root_dir, topdown=False))

    empty_dirs = set()
    total_files = removed_files = 0

    print(f"Walking in the root directory {str(root_dir)}...")

    for root, dirs, files in dir_walk:
        is_file = False

        for file in files:
            if path.splitext(file)[1][1:] not in keep_extensions or any(
                [phrase in str(file) for phrase in must_remove_phrases]
            ):
                (Path(root) / file).unlink()
                removed_files += 1

            else:
                is_file = True

            total_files += 1

        if not is_file:
            is_subdir = False

            for dir in dirs:
                if len(listdir(Path(root) / dir)) > 0:
                    is_subdir = True

            if not is_subdir or any(
                [phrase in str(root) for phrase in must_remove_phrases]
            ):
                empty_dirs.add(Path(root).absolute())

        dir_walk.update()

    print("Removing empty directories...")
    for empty_dir in tqdm(empty_dirs):
        try:
            rmtree(empty_dir)

        except:
            continue

    print(
        f"Total {formatter((total_files-removed_files), color='g', bold=True)}/{formatter(total_files, color='g', bold=True)} belong to {keep_extensions} extensions."
    )


def index_files(
    root_dir: Union[Path, str],
    indexed_repositories: Dict[str, dict],
    readme_save_dir: Union[Path, str],
    extensions: List[str] = ["py"],
) -> None:
    dir_walk = tqdm(walk(Path(root_dir), topdown=False))
    files_to_index = []

    for root, _, files in dir_walk:
        repository = None
        for part in Path(root).parts:
            if re.search("___", part):
                repository = indexed_repositories[part]

        for file in files:
            if path.splitext(file)[1][1:] in extensions:
                files_to_index.append(
                    download_readme(
                        {
                            "name": file,
                            "path": str(Path(root) / file),
                            "category": Path(root_dir).name,
                            "readme_url": repository["readme_url"],
                            "repository_name": repository["name"],
                            "repository_owner": repository["owner"],
                            "topics": repository["topics"],
                        },
                        readme_save_dir,
                    )
                )
            dir_walk.update()

    print(
        formatter(
            f"Indexed files for {formatter(Path(root_dir).name, color='g', bold=True)} ",
            tick=True,
        )
    )

    return files_to_index


def index_files_from_root(
    root_dir: Union[Path, str],
    save_path: Union[Path, str],
    readme_save_dir: Union[Path, str],
    repository_index_path: Union[Path, str],
    keywords: List[str],
) -> None:

    indexed_files = []

    with Path(repository_index_path).open() as fl:
        indexed_repositories = json.load(fl)

    for keyword in keywords:
        indexed_files += index_files(
            Path(root_dir) / keyword, indexed_repositories, readme_save_dir
        )

    # in case the directory to store the index file doesn't exist
    Path(save_path).parent.mkdir(parents=True, exist_ok=True)

    with open(save_path, "w") as fl:
        json.dump(indexed_files, fl)

    print(
        formatter(
            f"Saved the file successfully to {formatter(save_path, color='g', bold=True)}",
            tick=True,
        )
    )


class unmark:
    """
    converts markdown string to a text string
    source: https://stackoverflow.com/a/54923798
    """

    def __init__(self):
        Markdown.output_formats["plain"] = self.unmark_element
        self.__md = Markdown(output_format="plain")
        self.__md.stripTopLevelTags = False

    @classmethod
    def unmark_element(cls, element, stream=None):
        if stream is None:
            stream = StringIO()
        if element.text:
            stream.write(element.text)
        for sub in element:
            cls.unmark_element(sub, stream)
        if element.tail:
            stream.write(element.tail)
        return stream.getvalue()

    def __call__(self, text):
        return self.__md.convert(text)


def download_readme(file_json, save_dir: Union[Path, str]) -> dict:
    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    owner = file_json["repository_owner"]
    name = file_json["repository_name"]
    readme_path = save_dir / f"{owner}__{name}.txt"

    if not readme_path.exists():

        url = file_json["readme_url"]
        c = requests.get(url).text
        c = str(unmark()(c))

        with readme_path.open("w") as fl:
            fl.write(c)

    file_json["readme_path"] = str(readme_path)
    return file_json


def remove_commments(
    code: str,
) -> str:

    single_line_pattern = r"#(.*?)\n"
    multi_line_pattern = r'"""(.*?)"""'

    match = re.search(single_line_pattern, code)
    while match != None:
        start, end = match.span()
        end -= 1

        code = code[:start:] + code[end::]

        match = re.search(single_line_pattern, code, re.M)

    match = re.search(multi_line_pattern, code, re.M)
    while match != None:
        start, end = match.span()
        end -= 1

        code = code[:start:] + code[end::]

        match = re.search(multi_line_pattern, code, re.M)

    return code


def sentence_tokenize_and_select_random(text: str, num_sentences: int = 2) -> str:

    sentences = list(filter(lambda x: len(x) > 0, text.split(".")))

    if num_sentences < len(sentences):
        return ".".join(sentences)

    return ".".join(choices(sentences, k=num_sentences))



from pycoder.utils import *
from pycoder.data.classes import Code
from pycoder.data.modules import CodeDataset
from pycoder.model.trainer import get_trainer
from pycoder.imports import random_split, Path, Union
from pycoder.model.transformer import get_tokenier, get_model, save_transformers


def run_training(cfg, checkpoint_path: Union[Path, str] = None):
    codes = Code.load_from_json(cfg.FILE_INDEX_JSON_PATH)
    print(
        f"loaded {len(codes)} code instances from",
        formatter(cfg.FILE_INDEX_JSON_PATH, "g", True),
    )
    tokenizer = get_tokenier(cfg.MODEL_NAME, cfg.SPECIAL_TOKENS)
    model = get_model(cfg.MODEL_NAME, tokenizer, cfg.SPECIAL_TOKENS)
    print(
        f"loaded tokenizer and model  for {formatter(cfg.MODEL_NAME, 'g', True)} model"
    )

    ds = CodeDataset(
        codes,
        tokenizer,
        cfg.CONTROL_TOKENS,
        cfg.MAX_LENGTH,
        cfg.NUM_DESCRIPTION_SENTENCES,
    )

    print("dataset loaded successfully")

    val_sz = int(len(ds) * cfg.VAL_PCT)
    train_sz = len(ds) - val_sz
    train_ds, val_ds = random_split(ds, [train_sz, val_sz])

    print("len(training dataset):", len(train_ds))
    print("len(validation dataset):", len(val_ds))

    trainer = get_trainer(model, tokenizer, train_ds, val_ds, cfg)

    if checkpoint_path:
        print(formatter("resuming training", "g", True))
        trainer.train(checkpoint_path)
    else:
        print(formatter("initializing training", "g", True))
        trainer.train()

    save_transformers(
        cfg.MODEL_PATH,
        cfg.TOKENIZER_PATH,
        trainer.model,
        trainer.tokenizer,
    )



from pycoder.utils import formatter
from pycoder.model.transformer import load_transformers, save_transformers
from pycoder.imports import (
    pipeline,
    Union,
    Path,
    List,
    GPT2LMHeadModel,
    AutoTokenizer,
    rmtree,
)


class CodeInference:
    def __init__(
        self,
        model_path: Union[Path, str],
        tokenizer_path: Union[Path, str],
        control_tokens: dict,
        max_length: int,
        cuda: bool = False,
        verbose: bool = False,
    ):
        check_load_from_model_hub(model_path, tokenizer_path)
        model, tokenizer = load_transformers(model_path, tokenizer_path, verbose)
        model.config.task_specific_params["text-generation"]["max_length"] = max_length

        self.coder = pipeline(
            "text-generation",
            model=model,
            tokenizer=tokenizer,
            config={"max_length": max_length},
            device=0 if cuda else -1,
        )

        self.control_tokens = control_tokens

    def __call__(
        self, topics: Union[List[str], str], description: str, code_prefix: str = ""
    ):
        """
        sends in input:
        <|TOP|>TOPICS<|DES|>DESCRIPTION<|CODE|>

        hopes for the output as:
        <|TOP|>TOPICS<|DES|>DESCRIPTION<|CODE|>CODE<|EOS|>
        """

        if isinstance(topics, list):
            topics = ",".join(topics)

        inp = (
            self.control_tokens["topics_token"]
            + topics
            + self.control_tokens["description_token"]
            + description
            + self.control_tokens["code_token"]
            + code_prefix
        )

        out = self.coder(inp)
        out = list(
            map(
                lambda x: x["generated_text"].split(self.control_tokens["code_token"])[
                    1
                ],
                out,
            )
        )

        return out


def check_load_from_model_hub(
    model_path: Union[Path, str], tokenizer_path: Union[Path, str]
):
    from pycoder.config import HF_HUB_NAME, CACHE_DIR

    model_path = Path(model_path)
    tokenizer_path = Path(tokenizer_path)

    if not model_path.exists() or not tokenizer_path.exists():
        print(
            formatter(
                "\nModel and Tokenizer being downloaded and saved from ModelHub (once only)...\n",
                color="g",
            )
        )

        model = GPT2LMHeadModel.from_pretrained(
            HF_HUB_NAME, cache_dir=CACHE_DIR / "tokenizer"
        )
        tokenizer = AutoTokenizer.from_pretrained(
            HF_HUB_NAME, cache_dir=CACHE_DIR / "model"
        )

        save_transformers(model_path, tokenizer_path, model, tokenizer, verbose=False)

        print(
            formatter("\nModel and Tokenizer saved.", color="g", bold=True, tick=True)
            + "\n"
        )

        rmtree(CACHE_DIR)



from pycoder.imports import TrainingArguments, Trainer


def get_trainer(
    model: "torch.nn.Module",
    tokenizer: "transformer.PreTrainedTokenizer",
    train_ds: "torch.utils.data.Dataset",
    val_ds: "torch.utils.data.Dataset",
    cfg,
) -> "transformer.Trainer":
    training_args = TrainingArguments(
        output_dir=cfg.CHECKPOINT_PATH,
        num_train_epochs=cfg.EPOCHS,
        per_device_train_batch_size=cfg.TRAIN_BATCHSIZE,
        per_device_eval_batch_size=cfg.TRAIN_BATCHSIZE,
        gradient_accumulation_steps=cfg.BATCH_UPDATE,
        evaluation_strategy="epoch",
        fp16=cfg.FP16,
        fp16_opt_level=cfg.APEX_OPT_LEVEL,
        warmup_steps=cfg.WARMUP_STEPS,
        learning_rate=cfg.LR,
        adam_epsilon=cfg.EPS,
        weight_decay=0.01,
        save_total_limit=1,
        load_best_model_at_end=True,
        run_name=cfg.RUN_NAME,
        report_to="wandb",
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_ds,
        eval_dataset=val_ds,
        tokenizer=tokenizer,
    )

    return trainer



from pycoder.imports import Path
from pycoder.config import BOLD, GREEN, RED, CROSS, END, TICK


def formatter(
    st: str,
    color: str = None,
    bold: bool = False,
    tick: bool = False,
    cross: bool = False,
    at_end: bool = True,
) -> str:
    st = str(st)
    if color:
        color = GREEN if color.lower() == "g" else RED
    else:
        color = ""

    bold = BOLD if bold else ""

    if tick:
        st = st + " " + TICK if at_end else TICK + " " + st

    if cross:
        st = st + " " + CROSS if at_end else CROSS + " " + st

    return color + bold + st + END


def update_version():
    version = input("Enter version number: ")
    assert (
        len(version.split(".")) == 3
    ), """version represents RELEASE_VERSION.DEV_VERSION.MODEL_VERSION"""

    version_file = Path(__file__).parent / "version.py"

    with version_file.open("w") as fl:
        fl.write(
            f"""\"\"\"version represents RELEASE_VERSION.DEV_VERSION.MODEL_VERSION\"\"\"\n__version__ = "{version}\"
            """
        )



from setuptools import setup, find_packages
from pycoder.version import __version__
from pathlib import Path
import sys


with open("README.md") as f:
    long_description = f.read()

setup(
    name="pycoder",
    version=__version__,
    packages=find_packages(),
    description="A package to generate Python code from given topics and description",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Himanshu Dutta",
    author_email="meet.himanshu.dutta@gmail.com",
    url="https://github.com/himanshu-dutta/pycoder",
    install_requires=[
        "torch==1.8.1",
        "typer==0.3.2",
        "transformers==4.6.1",
        "markdown==3.3.4",
        "fastapi==0.65.1",
        "uvicorn==0.13.4",
        "pygments==2.9.0",
        "click-spinner==0.1.10",
    ],
    extras_require={"dev": ["pytest", "black", "wandb", "pygithub"]},
    entry_points={
        "console_scripts": [
            "pycoder=pycoder.api.cli:cli",
        ],
    },
    python_requires=">3.7",
    license="MIT License",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    platforms=["linux", "unix"],
)



from pycoder.api.main import query
from pycoder.imports import FastAPI, create_task, run

app = FastAPI()


async def load_model_call():
    query(
        topics="python",
        description="a sample call to load the model",
        verbose=True,
    )


@app.get("/query-code")
async def query_code(
    topics: str, description: str, prefix: str = "", max_length: int = 200
):
    result = query(topics, description, prefix, max_length, verbose=True)

    return {"code": result}


@app.get("/load-model")
async def load_model():
    create_task(load_model_call())
    return {"status": "loading model..."}


def main():
    run(app, host="127.0.0.1", log_level="info")



from pycoder.imports import json, Path, List, Union, dataclass, field, codecs


@dataclass
class Code:
    path: Path
    category: str
    readme_path: Path
    topics: List[str] = field(default_factory=list)
    language: str = "Python"

    def __post_init__(self):
        self._readme = None
        self._content = None

    @classmethod
    def load_from_raw(cls, instance: dict) -> "Code":
        return cls(
            Path(instance.get("path")),
            instance.get("category"),
            Path(instance.get("readme_path")),
            instance.get("topics"),
            instance.get("language"),
        )

    @classmethod
    def load_from_json(cls, json_path: Union[Path, str]) -> List["Code"]:
        with open(json_path, "r") as fl:
            metadata = json.load(fl)

        return list(filter(lambda x: x.is_valid(), map(cls.load_from_raw, metadata)))

    @property
    def content(self) -> str:
        try:
            if not self._content:
                with codecs.open(
                    self.path, "r", encoding="utf-8", errors="ignore"
                ) as fl:
                    self._content = fl.read()
            return self._content
        except:
            return None

    @property
    def readme(self) -> str:
        try:
            if not self._readme:
                with codecs.open(
                    self.readme_path, "r", encoding="utf-8", errors="ignore"
                ) as fl:
                    self._readme = fl.read()
            return self._readme
        except:
            return None

    def is_valid(self) -> bool:
        return bool(self.content and self.readme)
